@@ -1,40 +1,34 @@
 package App::fatten;
 
-our $DATE = '2015-01-05'; # DATE
-our $VERSION = '0.28'; # VERSION
+our $DATE = '2015-03-24'; # DATE
+our $VERSION = '0.29'; # VERSION
 
 use 5.010001;
 use strict;
 use warnings;
 use experimental 'smartmatch';
 use Log::Any '$log';
+use Log::Any::For::Builtins qw(system my_qx);
 BEGIN { no warnings; $main::Log_Level = 'info' }
 
-use App::tracepm;
-use Cwd qw(abs_path);
-use Dist::Util qw(list_dist_modules);
+use App::tracepm ();
+
 use File::chdir;
-use File::Copy;
-use File::Path qw(make_path remove_tree);
-use File::Spec;
-use File::Slurp::Tiny qw(write_file read_file);
-use File::Temp qw(tempfile tempdir);
-use List::MoreUtils qw(uniq);
-use List::Util qw(first);
-use Log::Any::For::Builtins qw(system my_qx);
-use Module::Path::More qw(module_path);
-use Proc::ChildError qw(explain_child_error);
-use File::MoreUtil qw(file_exists);
-use String::ShellQuote;
+use File::Slurper qw(write_text read_text);
 use version;
 
-sub _sq { shell_quote($_[0]) }
+sub _sq {
+    require String::ShellQuote;
+    String::ShellQuote::shell_quote($_[0]);
+}
 
 our %SPEC;
 
 sub _trace {
     my $self = shift;
 
+    return if $self->{trace_method} eq 'none';
+
     $log->debugf("  Tracing with method '%s' ...", $self->{trace_method});
     my $res = App::tracepm::tracepm(
         method => $self->{trace_method},
@@ -53,6 +47,13 @@ sub _trace {
 }
 
 sub _build_lib {
+    require Dist::Util;
+    require File::Copy;
+    require File::Find;
+    require File::Path;
+    require List::MoreUtils;
+    require Module::Path::More;
+
     my $self = shift;
 
     my $tempdir = $self->{tempdir};
@@ -60,39 +61,53 @@ sub _build_lib {
     my $totsize = 0;
     my $totfiles = 0;
 
-    my @mods; # modules to add
+    my %mods; # modules to add, key=name, val=path
 
     my $deps = $self->{deps};
     for (@$deps) {
         next if $_->{is_core} && $self->{exclude_core};
         next if $_->{is_xs};
         $log->debugf("  Adding module: %s (traced)", $_->{module});
-        push @mods, $_->{module};
+        $mods{$_->{module}} = undef;
     }
 
     for (@{ $self->{include} // [] }) {
         $log->debugf("  Adding module: %s (included)", $_);
-        push @mods, $_;
+        $mods{$_} = undef;
     }
 
     for (@{ $self->{include_dist} // [] }) {
-        my @distmods = list_dist_modules($_);
+        my @distmods = Dist::Util::list_dist_modules($_);
         if (@distmods) {
             $log->debugf("  Adding modules: %s (included dist)", join(", ", @distmods));
-            push @mods, @distmods;
+            $mods{$_} = undef for @distmods;
         } else {
             $log->debugf("  Adding module: %s (included dist, but can't find other modules)", $_);
-            push @mods, $_;
+            $mods{$_} = undef;
         }
     }
 
-    @mods = uniq(@mods);
+    for (@{ $self->{include_from} // [] }) {
+        $log->debugf("  Adding modules found in: %s", $_);
+        local $CWD = $_;
+        File::Find::find(
+            sub {
+                return unless -f;
+                return unless /\.pm$/i;
+                my $mod = $File::Find::dir eq '.' ? $_ : "$File::Find::dir/$_";
+                $mod =~ s!^\.[/\\]!!;
+                $mod =~ s![/\\]!::!g; $mod =~ s/\.pm$//i;
+                $log->debugf("    Adding module: %s", $mod);
+                $mods{$mod} = "$CWD/$_";
+            }, ".",
+        );
+    }
 
     # filter excluded
     my $excluded_distmods;
-    my @fmods;
+    my %fmods; # filtered mods
   MOD:
-    for my $mod (@mods) {
+    for my $mod (sort keys %mods) {
         if ($self->{exclude} && $mod ~~ @{ $self->{exclude} }) {
             $log->infof("Excluding %s: skipped", $mod);
             next MOD;
@@ -107,7 +122,7 @@ sub _build_lib {
             if (!$excluded_distmods) {
                 $excluded_distmods = [];
                 for (@{ $self->{exclude_dist} }) {
-                    push @$excluded_distmods, list_dist_modules($_);
+                    push @$excluded_distmods, Dist::Util::list_dist_modules($_);
                 }
             }
             if ($mod ~~ @$excluded_distmods) {
@@ -115,19 +130,20 @@ sub _build_lib {
                 next MOD;
             }
         }
-        push @fmods, $mod;
+        $fmods{$mod} = $mods{$mod};
     }
-    @mods = @fmods;
+    %mods = %fmods;
 
-    for my $mod (@mods) {
-        my $mpath = module_path(module=>$mod) or die "Can't find path for $mod";
+    for my $mod (sort keys %mods) {
+        my $mpath = $mods{$mod} // Module::Path::More::module_path(module=>$mod);
+        defined $mpath or die "Can't find path for $mod";
 
         my $modp = $mod; $modp =~ s!::!/!g; $modp .= ".pm";
         my ($dir) = $modp =~ m!(.+)/(.+)!;
         if ($dir) {
             my $dir_to_make = "$tempdir/lib/$dir";
             unless (-d $dir_to_make) {
-                make_path($dir_to_make) or die "Can't make_path: $dir_to_make";
+                File::Path::make_path($dir_to_make) or die "Can't make_path: $dir_to_make";
             }
         }
 
@@ -143,16 +159,16 @@ sub _build_lib {
                 );
             };
             $log->debug("  Stripping $mpath --> $modp ...");
-            my $src = read_file($mpath);
+            my $src = read_text($mpath);
             my $stripped = $stripper->strip($src);
-            write_file("$tempdir/lib/$modp", $stripped);
+            write_text("$tempdir/lib/$modp", $stripped);
         } elsif ($self->{strip}) {
             require Perl::Strip;
             my $strip = Perl::Strip->new;
             $log->debug("  Stripping $mpath --> $modp ...");
-            my $src = read_file($mpath);
+            my $src = read_text($mpath);
             my $stripped = $strip->strip($src);
-            write_file("$tempdir/lib/$modp", $stripped);
+            write_text("$tempdir/lib/$modp", $stripped);
         } elsif ($self->{squish}) {
             $log->debug("  Squishing $mpath --> $modp ...");
             require Perl::Squish;
@@ -160,7 +176,7 @@ sub _build_lib {
             $squish->file($mpath, "$tempdir/lib/$modp");
         } else {
             $log->debug("  Copying $mpath --> $tempdir/lib/$modp ...");
-            copy($mpath, "$tempdir/lib/$modp");
+            File::Copy::copy($mpath, "$tempdir/lib/$modp");
         }
 
         $totfiles++;
@@ -170,6 +186,8 @@ sub _build_lib {
 }
 
 sub _pack {
+    require Proc::ChildError;
+
     my $self = shift;
 
     my $tempdir = $self->{tempdir};
@@ -181,19 +199,19 @@ sub _pack {
         _sq($self->{abs_input_file}), " > ",
         _sq($self->{abs_output_file}),
     );
-    die "Can't fatpack file: ".explain_child_error() if $?;
+    die "Can't fatpack file: ".Proc::ChildError::explain_child_error() if $?;
 
     chmod 0755, $self->{abs_output_file};
 
     # replace shebang line (which contains perl path used by fatpack) with a
     # default system perl. perhaps make this configurable in the future.
     {
-        my $ct = read_file($self->{abs_output_file});
+        my $ct = read_text($self->{abs_output_file});
         my $shebang = $self->{shebang} // '#!/usr/bin/perl';
         $shebang = "#!$shebang" unless $shebang =~ /^#!/;
         $shebang =~ s/\R+//g;
         $ct =~ s{\A#!(.+)}{$shebang};
-        write_file($self->{abs_output_file}, $ct);
+        write_text($self->{abs_output_file}, $ct);
     }
 
     $log->infof("  Produced %s (%.1f KB)",
@@ -262,14 +280,16 @@ here.
 
 _
             schema => ['array*' => of => 'str*'],
-            cmdline_aliases => { I => {} },
+            cmdline_aliases => { I=>{} },
             tags => ['category:module-selection'],
-            element_completion => sub {
-                require Complete::Module;
-                my %args = @_;
-                Complete::Module::complete_module(word=>$args{word});
-            },
-            'x.schema.entity' => 'modulename',
+            'x.schema.element_entity' => 'modulename',
+        },
+        include_from => {
+            summary => 'Include extra modules under directories',
+            'summary.alt.numnoun.singular' => 'Include extra modules under a directory',
+            schema => ['array*' => of => 'str*'],
+            tags => ['category:module-selection'],
+            'x.schema.element_entity' => 'dirname',
         },
         include_dist => {
             summary => 'Include all modules of dist',
@@ -283,9 +303,7 @@ _
             schema => ['array*' => of => 'str*'],
             cmdline_aliases => {},
             tags => ['category:module-selection'],
-            element_completion => {
-            },
-            'x.schema.entity' => 'modulename',
+            'x.schema.element_entity' => 'distname',
         },
         exclude => {
             summary => 'Modules to exclude',
@@ -298,12 +316,7 @@ _
             schema => ['array*' => of => 'str*'],
             cmdline_aliases => { E => {} },
             tags => ['category:module-selection'],
-            element_completion => sub {
-                require Complete::Module;
-                my %args = @_;
-                Complete::Module::complete_module(word=>$args{word});
-            },
-            'x.schema.entity' => 'modulename',
+            'x.schema.element_entity' => 'modulename',
         },
         exclude_pattern => {
             summary => 'Regex patterns of modules to exclude',
@@ -316,7 +329,7 @@ _
             schema => ['array*' => of => 'str*'],
             cmdline_aliases => { p => {} },
             tags => ['category:module-selection'],
-            'x.schema.entity' => 'modulename',
+            #'x.schema.element_entity' => 'regex',
         },
         exclude_dist => {
             summary => 'Exclude all modules of dist',
@@ -330,12 +343,7 @@ _
             schema => ['array*' => of => 'str*'],
             cmdline_aliases => {},
             tags => ['category:module-selection'],
-            element_completion => sub {
-                require Complete::Dist;
-                my %args = @_;
-                Complete::Dist::complete_dist(word=>$args{word});
-            },
-            'x.schema.entity' => 'modulename',
+            'x.schema.element_entity' => 'distname',
         },
         exclude_core => {
             summary => 'Exclude core modules',
@@ -366,7 +374,7 @@ _
             summary => "Which method to use to trace dependencies",
             schema => ['str*', {
                 default => 'fatpacker',
-                in=>$trace_methods,
+                in=>[@$trace_methods, 'none'],
             }],
             description => <<'_',
 
@@ -375,6 +383,9 @@ Different tracing methods have different pro's and con's, one method might
 detect required modules that another method does not, and vice versa. There are
 several methods available, please see `App::tracepm` for more details.
 
+A special value of `none` is also provided. If this is selected, then fatten
+will not perform any tracing. Usually used in conjunction with `--include-from`.
+
 _
             cmdline_aliases => { t=>{} },
             tags => ['category:module-selection'],
@@ -390,12 +401,7 @@ Will be passed to the tracer. Will currently only affect the `fatpacker` and
 
 _
             tags => ['category:module-selection'],
-            element_completion => sub {
-                require Complete::Module;
-                my %args = @_;
-                Complete::Module::complete_module(word=>$args{word});
-            },
-            'x.schema.entity' => 'modulename',
+            'x.schema.element_entity' => 'modulename',
         },
         args => {
             summary => 'Script arguments',
@@ -511,10 +517,15 @@ _
     },
 };
 sub fatten {
+    require Cwd;
+    require File::MoreUtil;
+    require File::Spec;
+    require File::Temp;
+
     my %args = @_;
     my $self = __PACKAGE__->new(%args);
 
-    my $tempdir = tempdir(CLEANUP => 0);
+    my $tempdir = File::Temp::tempdir(CLEANUP => 0);
     $log->debugf("Created tempdir %s", $tempdir);
     $self->{tempdir} = $tempdir;
 
@@ -541,7 +552,7 @@ sub fatten {
 
     if ($self->{input_file} eq '-') {
         $self->{input_file_is_stdin} = 1;
-        $self->{input_file} = $self->{abs_input_file} = (tempfile())[1];
+        $self->{input_file} = $self->{abs_input_file} = (File::Temp::tempfile())[1];
         open my($fh), ">", $self->{abs_input_file}
             or return [500, "Can't write temporary input file '$self->{abs_input_file}': $!"];
         local $_; while (<STDIN>) { print $fh $_ }
@@ -549,7 +560,7 @@ sub fatten {
     } else {
         (-f $self->{input_file})
             or return [500, "No such input file: $self->{input_file}"];
-        $self->{abs_input_file} = abs_path($self->{input_file}) or return
+        $self->{abs_input_file} = Cwd::abs_path($self->{input_file}) or return
             [500, "Can't find absolute path of input file $self->{input_file}"];
     }
 
@@ -559,11 +570,11 @@ sub fatten {
         if (defined $output_file) {
             if ($output_file eq '-') {
                 $self->{output_file_is_stdout} = 1;
-                $self->{output_file} = $self->{abs_output_file} = (tempfile())[1];
+                $self->{output_file} = $self->{abs_output_file} = (File::Temp::tempfile())[1];
                 last;
             } else {
                 return [412, "Output file '$output_file' exists, won't overwrite (see --overwrite)"]
-                    if file_exists($output_file) && !$self->{overwrite};
+                    if File::MoreUtil::file_exists($output_file) && !$self->{overwrite};
                 last if open my($fh), ">", $output_file;
                 return [500, "Can't write to output file '$output_file': $!"];
             }
@@ -575,13 +586,13 @@ sub fatten {
         # try <input>.fatpack in the source directory
         $output_file = File::Spec->catpath($vol, $dir, "$file.fatpack");
         return [412, "Output file '$output_file' exists, won't overwrite (see --overwrite)"]
-            if file_exists($output_file) && !$self->{overwrite};
+            if File::MoreUtil::file_exists($output_file) && !$self->{overwrite};
         last if open $fh, ">", $output_file;
 
         # if failed, try <input>.fatpack in the current directory
         $output_file = "$CWD/$file.fatpack";
         return [412, "Output file '$output_file' exists, won't overwrite (see --overwrite)"]
-            if file_exists($output_file) && !$self->{overwrite};
+            if File::MoreUtil::file_exists($output_file) && !$self->{overwrite};
         last if open $fh, ">", $output_file;
 
         # failed too, bail
@@ -589,7 +600,7 @@ sub fatten {
                     "current directory: $!"];
     }
     $self->{output_file} = $output_file;
-    $self->{abs_output_file} //= abs_path($output_file) or return
+    $self->{abs_output_file} //= Cwd::abs_path($output_file) or return
         [500, "Can't find absolute path of output file '$self->{output_file}'"];
 
     $log->infof("Tracing dependencies ...");
@@ -605,7 +616,7 @@ sub fatten {
         $log->infof("Keeping tempdir %s for debugging", $tempdir);
     } else {
         $log->debugf("Deleting tempdir %s ...", $tempdir);
-        remove_tree($tempdir);
+        File::Path::remove_tree($tempdir);
     }
 
     if ($self->{input_file_is_stdin}) {
@@ -636,7 +647,7 @@ App::fatten - Pack your dependencies onto your script file
 
 =head1 VERSION
 
-This document describes version 0.28 of App::fatten (from Perl distribution App-fatten), released on 2015-01-05.
+This document describes version 0.29 of App::fatten (from Perl distribution App-fatten), released on 2015-03-24.
 
 =head1 SYNOPSIS
 
@@ -716,6 +727,10 @@ Just like the C<include> option, but will include module as well as other module
 from the same distribution. Module name must be the main module of the
 distribution. Will determine other modules from the C<.packlist> file.
 
+=item * B<include_from> => I<array[str]>
+
+Include extra modules under directories.
+
 =item * B<input_file> => I<str> (default: "-")
 
 Path to input file (script to be fatpacked).
@@ -801,6 +816,9 @@ Different tracing methods have different pro's and con's, one method might
 detect required modules that another method does not, and vice versa. There are
 several methods available, please see C<App::tracepm> for more details.
 
+A special value of C<none> is also provided. If this is selected, then fatten
+will not perform any tracing. Usually used in conjunction with C<--include-from>.
+
 =item * B<use> => I<array[str]>
 
 Additional modules to "use".
@@ -829,7 +847,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/App-fatten
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/sharyanto/perl-App-fatten>.
+Source repository is at L<https://github.com/perlancar/perl-App-fatten>.
 
 =head1 BUGS
 
